import unittest

import analogai.foundation.extensions.belief_query.tests.env_setup
from analogai.foundation.core.value_objects.relation import Relation
from analogai.foundation.core.value_objects.cause import Cause
from analogai.foundation.core.value_objects.effect import Effect
from analogai.foundation.common.dtos.notion_expression import NotionExpression
from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.core.value_objects.attribute import Attribute
from analogai.foundation.common.enums.criterion import Criterion
from analogai.foundation.common.enums.certainty import Certainty
from analogai.foundation.setup.factories import BeliefQueryServiceFactory
from analogai.foundation.tests.functional.services.setup import ServicesTestConfig


def create_learning(
    learnings_service,
    user_id,
    agent_id,
    subject_notion,
    result_criteria,
    modifier=None,
    probability=None,
    validity=None,
    statement_service=None,
    notion_service=None,
    belief_service=None,
):
    relation = result_criteria.get(Criterion.Relation)
    object_name = result_criteria.get(Criterion.ObjectCaption)
    if not relation or not object_name:
        raise ValueError("Belief result requires Relation and ObjectCaption")
    if not isinstance(relation, Relation):
        raise TypeError("Criterion.Relation must be a Relation instance")
    if statement_service is None or notion_service is None:
        raise ValueError("statement_service and notion_service are required to create learning statements")
    if belief_service is None:
        belief_service = learnings_service._belief_service
    attributes = None
    attr_values = result_criteria.get(Criterion.AttributeValue)
    if attr_values:
        attributes = []
        for item in attr_values:
            if isinstance(item, Attribute):
                attributes.append(item)
                continue
            if isinstance(item, dict):
                tag = item.get("tag") or item.get("kind")
                value = item.get("value") or item.get("min_value") or item.get("max_value")
                unit = item.get("unit")
                if tag is not None and value is not None:
                    attributes.append(
                        Attribute(tag=str(tag), value=value, unit=str(unit) if unit else None)
                    )
    complement_notion = notion_service.find_or_create(
        user_id,
        agent_id,
        NotionExpression(caption=str(object_name), attributes=attributes),
    )
    relation_obj = relation
    prob = probability if probability is not None else result_criteria.get(Criterion.RelationProbability, 1.0)
    valid = validity if validity is not None else 1.0
    final_modifier = modifier
    statement_service.create(
        user_id=user_id,
        agent_id=agent_id,
        subject_notion=subject_notion,
        complement_notion=complement_notion,
        relation=relation_obj,
        probability=prob,
        validity=valid,
        modifier=final_modifier,
    )
    return belief_service.find(
        subject_notion=subject_notion,
        complement_notion=complement_notion,
        relation=relation_obj,
        modifier=final_modifier,
    )


class TestBeliefQueryService(unittest.TestCase):
    def setUp(self):
        self.config = ServicesTestConfig()
        self.notion_service = self.config.notion_service
        self.statement_service = self.config.statement_service
        self.belief_service = self.config.belief_service
        self.hypothetical_statement_service = self.config.hypothetical_statement_service
        self.service = BeliefQueryServiceFactory(
            belief_service=self.belief_service,
        ).get_service()
        self.user_id = self.config.user.id
        self.agent_id = self.config.agent.id
        self.subject_notion = self.notion_service.find_or_create(
            self.user_id, self.agent_id, NotionExpression(caption="socrates")
        )

    def _search(self):
        return self.service.search(self.user_id, self.agent_id).exactModifierMatch()

    def tearDown(self):
        self.config.tear_down()

    def test_create_belief_returns_learning_with_dynamic_resolution(self):
        result_criteria = {
            Criterion.Relation: Relation.from_string("is"),
            Criterion.ObjectCaption: "human",
        }
        belief = create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            result_criteria,
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        self.assertEqual(belief.subject_notion.caption, "socrates")
        self.assertEqual(belief.complement_notion.caption, "human")
        self.assertEqual(belief.relation, Relation.from_string("is"))

    def test_create_statement_returns_learning_with_dynamic_resolution(self):
        result_criteria = {
            Criterion.Relation: Relation.from_string("sells"),
            Criterion.ObjectCaption: "iphone",
        }
        belief = create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            result_criteria,
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        self.assertEqual(belief.subject_notion.caption, "socrates")
        self.assertEqual(belief.relation, Relation.from_string("sells"))
        self.assertEqual(belief.complement_notion.caption, "iphone")

    def test_search_includes_categorical_deductions(self):
        human = self.notion_service.find_or_create(
            self.user_id, self.agent_id, NotionExpression(caption="human")
        )
        mortal = self.notion_service.find_or_create(
            self.user_id, self.agent_id, NotionExpression(caption="mortal")
        )
        self.statement_service.create(
            user_id=self.user_id,
            agent_id=self.agent_id,
            subject_notion=self.subject_notion,
            complement_notion=human,
            relation=Relation.from_string("is"),
            probability=1.0,
            validity=1.0,
        )
        self.statement_service.create(
            user_id=self.user_id,
            agent_id=self.agent_id,
            subject_notion=human,
            complement_notion=mortal,
            relation=Relation.from_string("is"),
            probability=1.0,
            validity=1.0,
        )

        results = self._search().where(
            Criterion.SubjectCaption, "socrates"
        ).where(
            Criterion.Relation, Relation.from_string("is")
        ).where(
            Criterion.ObjectCaption, "mortal"
        ).list()

        self.assertEqual(len(results), 1)

    def test_search_includes_categorical_and_conditional_deductions(self):
        human = self.notion_service.find_or_create(
            self.user_id, self.agent_id, NotionExpression(caption="human")
        )
        mortal = self.notion_service.find_or_create(
            self.user_id, self.agent_id, NotionExpression(caption="mortal")
        )
        self.statement_service.create(
            user_id=self.user_id,
            agent_id=self.agent_id,
            subject_notion=self.subject_notion,
            complement_notion=human,
            relation=Relation.from_string("is"),
            probability=1.0,
            validity=1.0,
        )
        self.statement_service.create(
            user_id=self.user_id,
            agent_id=self.agent_id,
            subject_notion=human,
            complement_notion=mortal,
            relation=Relation.from_string("is"),
            probability=1.0,
            validity=1.0,
        )
        self.hypothetical_statement_service.create(
            user_id=self.user_id,
            agent_id=self.agent_id,
            causes=[
                Cause({
                    Criterion.SubjectCaption: "socrates",
                    Criterion.ObjectCaption: "human",
                    Criterion.Relation: Relation.from_string("is"),
                })
            ],
            effect=Effect({
                Criterion.SubjectCaption: "socrates",
                Criterion.ObjectCaption: "wisdom",
                Criterion.Relation: Relation.from_string("causes"),
                Criterion.Certainty: Certainty.UNCERTAIN,
            }),
        )

        categorical_results = self._search().where(
            Criterion.SubjectCaption, "socrates"
        ).where(
            Criterion.Relation, Relation.from_string("is")
        ).where(
            Criterion.ObjectCaption, "mortal"
        ).list()

        conditional_results = self._search().where(
            Criterion.SubjectCaption, "socrates"
        ).where(
            Criterion.Relation, Relation.from_string("causes")
        ).where(
            Criterion.ObjectCaption, "wisdom"
        ).list()

        self.assertEqual(len(categorical_results), 1)
        self.assertEqual(len(conditional_results), 1)


if __name__ == "__main__":
    unittest.main()


import unittest

import analogai.foundation.extensions.belief_query.tests.env_setup
from analogai.foundation.common.dtos.notion_expression import NotionExpression
from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.core.value_objects.attribute import Attribute
from analogai.foundation.core.value_objects.time import Time
from analogai.foundation.core.value_objects.relation import Relation
from analogai.foundation.common.enums.deontic_modality import DeonticModality
from analogai.foundation.common.enums.criterion import Criterion
from analogai.foundation.setup.factories import BeliefQueryServiceFactory
from analogai.foundation.tests.functional.services.setup import ServicesTestConfig


def create_learning(
    learnings_service,
    user_id,
    agent_id,
    subject_notion,
    result_criteria,
    modifier=None,
    probability=None,
    validity=None,
    statement_service=None,
    notion_service=None,
    belief_service=None,
):
    relation = result_criteria.get(Criterion.Relation)
    object_name = result_criteria.get(Criterion.ObjectCaption)
    if not relation or not object_name:
        raise ValueError("Belief result requires Relation and ObjectCaption")
    if not isinstance(relation, Relation):
        raise TypeError("Criterion.Relation must be a Relation instance")
    if statement_service is None or notion_service is None:
        raise ValueError("statement_service and notion_service are required to create learning statements")
    if belief_service is None:
        belief_service = learnings_service._belief_service
    attributes = None
    attr_values = result_criteria.get(Criterion.AttributeValue)
    if attr_values:
        attributes = []
        for item in attr_values:
            if isinstance(item, Attribute):
                attributes.append(item)
                continue
            if isinstance(item, dict):
                tag = item.get("tag") or item.get("kind")
                value = item.get("value") or item.get("min_value") or item.get("max_value")
                unit = item.get("unit")
                if tag is not None and value is not None:
                    attributes.append(
                        Attribute(tag=str(tag), value=value, unit=str(unit) if unit else None)
                    )
    complement_notion = notion_service.find_or_create(
        user_id,
        agent_id,
        NotionExpression(caption=str(object_name), attributes=attributes),
    )
    relation_obj = relation
    prob = probability if probability is not None else result_criteria.get(Criterion.RelationProbability, 1.0)
    valid = validity if validity is not None else 1.0
    final_modifier = modifier
    statement_service.create(
        user_id=user_id,
        agent_id=agent_id,
        subject_notion=subject_notion,
        complement_notion=complement_notion,
        relation=relation_obj,
        probability=prob,
        validity=valid,
        modifier=final_modifier,
    )
    return belief_service.find(
        subject_notion=subject_notion,
        complement_notion=complement_notion,
        relation=relation_obj,
        modifier=final_modifier,
    )


class TestBeliefSearch(unittest.TestCase):
    def setUp(self):
        self.config = ServicesTestConfig()
        self.notion_service = self.config.notion_service
        self.statement_service = self.config.statement_service
        self.belief_service = self.config.belief_service
        self.service = BeliefQueryServiceFactory(
            belief_service=self.belief_service,
        ).get_service()
        self.user_id = self.config.user.id
        self.agent_id = self.config.agent.id
        self.subject_notion = self.notion_service.find_or_create(
            self.user_id, self.agent_id, NotionExpression(caption="socrates")
        )

    def _search(self):
        return self.service.search(self.user_id, self.agent_id).exactModifierMatch()

    def tearDown(self):
        self.config.tear_down()

    def test_search_exists_returns_true_when_learning_matches(self):
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            {
                Criterion.Relation: Relation.from_string("sells"),
                Criterion.ObjectCaption: "iphone",
            },
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        result = self._search().where(
            Criterion.Relation, Relation.from_string("sells")
        ).exists()
        self.assertTrue(result)

    def test_search_exists_returns_false_when_no_match(self):
        result = self._search().where(
            Criterion.Relation, Relation.from_string("nonexistent")
        ).exists()
        self.assertFalse(result)

    def test_search_list_returns_matching_learnings(self):
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            {
                Criterion.Relation: Relation.from_string("sells"),
                Criterion.ObjectCaption: "iphone",
            },
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        learnings = self._search().where(
            Criterion.SubjectCaption, "socrates"
        ).list()
        self.assertEqual(len(learnings), 1)
        self.assertEqual(learnings[0].relation, Relation.from_string("sells"))

    def test_search_count_returns_correct_count(self):
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            {
                Criterion.Relation: Relation.from_string("sells"),
                Criterion.ObjectCaption: "iphone",
            },
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            {
                Criterion.Relation: Relation.from_string("is"),
                Criterion.ObjectCaption: "human",
            },
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        count = self._search().count()
        self.assertEqual(count, 2)

    def test_search_has_time_filters_by_time_existence(self):
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            {
                Criterion.Relation: Relation.from_string("sells"),
                Criterion.ObjectCaption: "iphone",
            },
            modifier=Modifier(from_time=Time(year=2024, month=1, day=15)),
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            {
                Criterion.Relation: Relation.from_string("sells"),
                Criterion.ObjectCaption: "android",
            },
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        with_time = self._search().whereNotNull(
            Criterion.Time
        ).list()
        without_time = self._search().whereNull(
            Criterion.Time
        ).list()
        self.assertEqual(len(with_time), 1)
        self.assertEqual(len(without_time), 1)
        self.assertEqual(with_time[0].complement_notion.caption, "iphone")
        self.assertEqual(without_time[0].complement_notion.caption, "android")

    def test_search_min_time_max_time_filter_by_time_range(self):
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            {
                Criterion.Relation: Relation.from_string("sells"),
                Criterion.ObjectCaption: "iphone",
            },
            modifier=Modifier(
                from_time=Time(year=2024, month=3, day=1),
                to_time=Time(year=2024, month=3, day=31),
            ),
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            {
                Criterion.Relation: Relation.from_string("sells"),
                Criterion.ObjectCaption: "android",
            },
            modifier=Modifier(from_time=Time(year=2024, month=1, day=1)),
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        in_range = self._search().whereMin(
            Criterion.Time, Time(year=2024, month=2, day=1)
        ).whereMax(Criterion.Time, Time(year=2024, month=4, day=1)).list()
        self.assertEqual(len(in_range), 1)
        self.assertEqual(in_range[0].complement_notion.caption, "iphone")

    def test_search_location_filters_by_modifier_location(self):
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            {
                Criterion.Relation: Relation.from_string("sells"),
                Criterion.ObjectCaption: "iphone",
            },
            modifier=Modifier(location="store"),
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            {
                Criterion.Relation: Relation.from_string("sells"),
                Criterion.ObjectCaption: "android",
            },
            modifier=Modifier(location="online"),
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        at_store = self._search().where(
            Criterion.Location, "store"
        ).list()
        self.assertEqual(len(at_store), 1)
        self.assertEqual(at_store[0].complement_notion.caption, "iphone")

    def test_search_has_deontic_modality_filters_by_deontic_existence(self):
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            {
                Criterion.Relation: Relation.from_string("sells"),
                Criterion.ObjectCaption: "iphone",
            },
            modifier=Modifier(deontic_modality=DeonticModality.OBLIGATORY),
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            {
                Criterion.Relation: Relation.from_string("sells"),
                Criterion.ObjectCaption: "android",
            },
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        with_deontic = self._search().whereNotNull(
            Criterion.DeonticModality
        ).list()
        without_deontic = self._search().whereNull(
            Criterion.DeonticModality
        ).list()
        self.assertEqual(len(with_deontic), 1)
        self.assertEqual(len(without_deontic), 1)
        self.assertEqual(with_deontic[0].complement_notion.caption, "iphone")
        self.assertEqual(without_deontic[0].complement_notion.caption, "android")

    def test_search_deontic_modality_filters_by_deontic_value(self):
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            {
                Criterion.Relation: Relation.from_string("sells"),
                Criterion.ObjectCaption: "iphone",
            },
            modifier=Modifier(deontic_modality=DeonticModality.OBLIGATORY),
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            {
                Criterion.Relation: Relation.from_string("sells"),
                Criterion.ObjectCaption: "android",
            },
            modifier=Modifier(deontic_modality=DeonticModality.PERMITTED),
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        obligatory = self._search().where(
            Criterion.DeonticModality, DeonticModality.OBLIGATORY
        ).list()
        permitted_by_str = self._search().where(
            Criterion.DeonticModality, "permitted"
        ).list()
        self.assertEqual(len(obligatory), 1)
        self.assertEqual(len(permitted_by_str), 1)
        self.assertEqual(obligatory[0].complement_notion.caption, "iphone")
        self.assertEqual(permitted_by_str[0].complement_notion.caption, "android")

    def test_search_measure_kind_filters_by_measure_kind(self):
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            {
                Criterion.Relation: Relation.from_string("buys"),
                Criterion.ObjectCaption: "phone",
                Criterion.AttributeValue: [
                    {"tag": "price", "value": 900, "unit": "usd"}
                ],
            },
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            {
                Criterion.Relation: Relation.from_string("buys"),
                Criterion.ObjectCaption: "book",
                Criterion.AttributeValue: [
                    {"tag": "weight", "value": 2, "unit": "kg"}
                ],
            },
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        priced = self._search().where(
            Criterion.AttributeTag, "price"
        ).list()
        self.assertEqual(len(priced), 1)
        self.assertEqual(priced[0].complement_notion.caption, "phone")

    def test_search_measure_number_whereMax_filters_by_numeric_measure(self):
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            {
                Criterion.Relation: Relation.from_string("buys"),
                Criterion.ObjectCaption: "phone",
                Criterion.AttributeValue: [
                    {"tag": "price", "value": 900, "unit": "usd"}
                ],
            },
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            {
                Criterion.Relation: Relation.from_string("buys"),
                Criterion.ObjectCaption: "case",
                Criterion.AttributeValue: [
                    {"tag": "price", "value": 30, "unit": "usd"}
                ],
            },
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        affordable = self._search().whereMax(
            Criterion.AttributeMaxValue, ("price", 100, "usd")
        ).list()
        self.assertEqual(len(affordable), 1)
        self.assertEqual(affordable[0].complement_notion.caption, "case")

    def test_search_or_where_matches_any_value(self):
        notion_service = self.notion_service
        plato = notion_service.find_or_create(
            self.user_id, self.agent_id, NotionExpression(caption="plato")
        )
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            {
                Criterion.Relation: Relation.from_string("sells"),
                Criterion.ObjectCaption: "iphone",
            },
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            plato,
            {
                Criterion.Relation: Relation.from_string("sells"),
                Criterion.ObjectCaption: "android",
            },
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        learnings = self._search().where(
            Criterion.SubjectCaption, "socrates"
        ).orWhere(Criterion.SubjectCaption, "plato").list()
        self.assertEqual(len(learnings), 2)
        subjects = {belief.subject_notion.caption for belief in learnings}
        self.assertEqual(subjects, {"socrates", "plato"})

    def test_search_or_where_with_different_criterion(self):
        notion_service = self.notion_service
        mortal = notion_service.find_or_create(
            self.user_id, self.agent_id, NotionExpression(caption="mortal")
        )
        plato = notion_service.find_or_create(
            self.user_id, self.agent_id, NotionExpression(caption="plato")
        )
        notion_service.find_or_create(
            self.user_id, self.agent_id, NotionExpression(caption="is")
        )
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            {
                Criterion.Relation: Relation.from_string("is"),
                Criterion.ObjectCaption: "mortal",
            },
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            plato,
            {
                Criterion.Relation: Relation.from_string("is"),
                Criterion.ObjectCaption: "socrates",
            },
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        learnings = self._search().where(
            Criterion.SubjectCaption, "socrates"
        ).orWhere(Criterion.ObjectCaption, "socrates").list()
        self.assertEqual(len(learnings), 2)
        subjects_and_objects = {(b.subject_notion.caption, b.complement_notion.caption) for b in learnings}
        self.assertEqual(subjects_and_objects, {("socrates", "mortal"), ("plato", "socrates")})

    def test_search_where_group_sets_main_clause(self):
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            {
                Criterion.Relation: Relation.from_string("sells"),
                Criterion.ObjectCaption: "iphone",
            },
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        learnings = self._search().whereGroup({
            Criterion.SubjectCaption: "socrates",
            Criterion.Relation: Relation.from_string("sells"),
        }).list()
        self.assertEqual(len(learnings), 1)
        self.assertEqual(learnings[0].complement_notion.caption, "iphone")

    def test_search_or_where_group_adds_alternative_clause(self):
        notion_service = self.notion_service
        plato = notion_service.find_or_create(
            self.user_id, self.agent_id, NotionExpression(caption="plato")
        )
        notion_service.find_or_create(
            self.user_id, self.agent_id, NotionExpression(caption="mortal")
        )
        notion_service.find_or_create(
            self.user_id, self.agent_id, NotionExpression(caption="is")
        )
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            {
                Criterion.Relation: Relation.from_string("is"),
                Criterion.ObjectCaption: "mortal",
            },
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            plato,
            {
                Criterion.Relation: Relation.from_string("is"),
                Criterion.ObjectCaption: "socrates",
            },
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        learnings = self._search().whereGroup({
            Criterion.SubjectCaption: "socrates",
            Criterion.ObjectCaption: "mortal",
        }).orWhereGroup({
            Criterion.SubjectCaption: "plato",
            Criterion.ObjectCaption: "socrates",
        }).list()
        self.assertEqual(len(learnings), 2)
        subjects_and_objects = {(b.subject_notion.caption, b.complement_notion.caption) for b in learnings}
        self.assertEqual(subjects_and_objects, {("socrates", "mortal"), ("plato", "socrates")})

    def test_search_where_in_matches_any_value_in_list(self):
        notion_service = self.notion_service
        plato = notion_service.find_or_create(
            self.user_id, self.agent_id, NotionExpression(caption="plato")
        )
        aristotle = notion_service.find_or_create(
            self.user_id, self.agent_id, NotionExpression(caption="aristotle")
        )
        notion_service.find_or_create(
            self.user_id, self.agent_id, NotionExpression(caption="sells")
        )
        notion_service.find_or_create(
            self.user_id, self.agent_id, NotionExpression(caption="iphone")
        )
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            {
                Criterion.Relation: Relation.from_string("sells"),
                Criterion.ObjectCaption: "iphone",
            },
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            plato,
            {
                Criterion.Relation: Relation.from_string("sells"),
                Criterion.ObjectCaption: "iphone",
            },
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            aristotle,
            {
                Criterion.Relation: Relation.from_string("sells"),
                Criterion.ObjectCaption: "iphone",
            },
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        learnings = self._search().whereIn(
            Criterion.SubjectCaption, ["socrates", "plato"]
        ).where(Criterion.Relation, Relation.from_string("sells")).list()
        self.assertEqual(len(learnings), 2)
        subjects = {b.subject_notion.caption for b in learnings}
        self.assertEqual(subjects, {"socrates", "plato"})

    def test_search_or_where_in_adds_alternative_with_list(self):
        notion_service = self.notion_service
        plato = notion_service.find_or_create(
            self.user_id, self.agent_id, NotionExpression(caption="plato")
        )
        notion_service.find_or_create(
            self.user_id, self.agent_id, NotionExpression(caption="sells")
        )
        notion_service.find_or_create(
            self.user_id, self.agent_id, NotionExpression(caption="iphone")
        )
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            self.subject_notion,
            {
                Criterion.Relation: Relation.from_string("sells"),
                Criterion.ObjectCaption: "iphone",
            },
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        create_learning(
            self.service,
            self.user_id,
            self.agent_id,
            plato,
            {
                Criterion.Relation: Relation.from_string("sells"),
                Criterion.ObjectCaption: "android",
            },
            statement_service=self.statement_service,
            notion_service=self.notion_service,
        )
        learnings = self._search().where(
            Criterion.SubjectCaption, "socrates"
        ).orWhereIn(Criterion.SubjectCaption, ["plato"]).list()
        self.assertEqual(len(learnings), 2)
        subjects_and_objects = {(b.subject_notion.caption, b.complement_notion.caption) for b in learnings}
        self.assertEqual(subjects_and_objects, {("socrates", "iphone"), ("plato", "android")})



if __name__ == "__main__":
    unittest.main()


import os
import unittest
from typing import List

from analogai.foundation import config
from analogai.foundation.common.dtos.notion_expression import NotionExpression
from analogai.foundation.common.enums.criterion import Criterion
from analogai.foundation.core.value_objects.relation import Relation
from analogai.foundation.setup.factories import BeliefQueryServiceFactory
from analogai.foundation.tests.functional.services.setup import ServicesTestConfig


class TestVectorSearchSimilarity(unittest.TestCase):
    def setUp(self):
        if os.getenv("USE_VECTOR_SEARCH", "false").lower() != "true":
            self.skipTest("USE_VECTOR_SEARCH is not enabled")

        self.config = ServicesTestConfig()
        self.notion_service = self.config.notion_service
        self.belief_service = self.config.belief_service
        self.statement_service = self.config.statement_service
        self.user = self.config.user
        self.agent = self.config.agent

    def tearDown(self):
        self.config.tear_down()

    def _expr(self, caption: str, context: List[NotionExpression] = None) -> NotionExpression:
        return NotionExpression(caption=caption, context=context or [])

    def test_notion_similarity_threshold_low_matches(self):
        created = self.notion_service.find_or_create(
            self.user.id,
            self.agent.id,
            self._expr("alpha"),
        )
        found = self.notion_service.find(
            self.agent.id,
            self._expr("beta"),
            similarity_threshold=0.0,
        )
        self.assertIsNotNone(found)
        self.assertEqual(found.id, created.id)

    def test_notion_similarity_threshold_high_rejects(self):
        self.notion_service.find_or_create(
            self.user.id,
            self.agent.id,
            self._expr("human"),
        )
        found = self.notion_service.find(
            self.agent.id,
            self._expr("banana"),
            similarity_threshold=0.99,
        )
        self.assertIsNone(found)

    def test_relation_similarity_respects_threshold(self):
        subject = self.notion_service.find_or_create(
            self.user.id,
            self.agent.id,
            self._expr("socrates"),
        )
        complement = self.notion_service.find_or_create(
            self.user.id,
            self.agent.id,
            self._expr("human"),
        )
        self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=subject,
            relation=Relation.from_string("is"),
            complement_notion=complement,
            probability=1.0,
            validity=1.0,
        )

        found_is = self.belief_service.find_by_expressions(
            agent_id=self.agent.id,
            subject_notion_expression=self._expr("socrates"),
            complement_notion_expression=self._expr("human"),
            relation=Relation.from_string("is"),
        )
        self.assertIsNotNone(found_is)

        found_causes = self.belief_service.find_by_expressions(
            agent_id=self.agent.id,
            subject_notion_expression=self._expr("socrates"),
            complement_notion_expression=self._expr("human"),
            relation=Relation.from_string("causes"),
        )
        self.assertIsNone(found_causes)

    def test_semantic_similarity_human_person(self):
        threshold_str = os.getenv("SEMANTIC_SIMILARITY_THRESHOLD", "0.4")
        try:
            threshold = float(threshold_str)
        except ValueError:
            threshold = 0.4

        created = self.notion_service.find_or_create(
            self.user.id,
            self.agent.id,
            self._expr("human"),
        )
        found = self.notion_service.find(
            self.agent.id,
            self._expr("person"),
            similarity_threshold=threshold,
        )
        self.assertIsNotNone(found)
        self.assertEqual(found.id, created.id)

    def test_belief_query_similarity_socrates_human_person(self):
        if config.SIMILARITY_CUTOFF > 0.8:
            self.skipTest("SIMILARITY_CUTOFF too high for semantic match in test")

        belief_query = BeliefQueryServiceFactory(
            belief_service=self.belief_service,
        ).get_service()

        subject = self.notion_service.find_or_create(
            self.user.id,
            self.agent.id,
            self._expr("socrates"),
        )
        complement = self.notion_service.find_or_create(
            self.user.id,
            self.agent.id,
            self._expr("human"),
        )
        self.statement_service.create(
            user_id=self.user.id,
            agent_id=self.agent.id,
            subject_notion=subject,
            relation=Relation.from_string("is"),
            complement_notion=complement,
            probability=1.0,
            validity=1.0,
        )

        results = belief_query.search(self.user.id, self.agent.id).where(
            Criterion.SubjectCaption,
            "socrates",
        ).where(
            Criterion.Relation,
            Relation.from_string("is"),
        ).where(
            Criterion.ObjectCaption,
            "person",
        ).list()

        self.assertTrue(results)


if __name__ == "__main__":
    unittest.main()



from typing import Dict, List, Optional, Any, TYPE_CHECKING, Iterable, Tuple
from datetime import datetime, timedelta

from analogai.foundation.modules.notion.notion_service import NotionService
from analogai.foundation.modules.direct_statement.statement_service_base import StatementServiceBase

if TYPE_CHECKING:
    from analogai.foundation.common.enums.criterion import Criterion
from analogai.foundation.modules.belief.belief import Belief
from analogai.foundation.modules.notion.notion import Notion
from analogai.foundation.modules.shared.statement_base import StatementBase
from analogai.foundation.modules.direct_statement.direct_statement import DirectStatement
from analogai.foundation.core.value_objects.modifier import Modifier
from analogai.foundation.core.value_objects.relation import Relation
from analogai.foundation.core.value_objects.time import Time
from analogai.foundation.core.value_objects.attribute import Attribute
from analogai.foundation.common.dtos.notion_expression import NotionExpression
from analogai.foundation.common.logging.app_logger import app_logger
from analogai.foundation.common.utils.modifier_utils.modifier_matcher import ModifierMatcher
from analogai.foundation.common.utils.statement_serializers import serialize_statement
from typeguard import typechecked


BeliefKey = Tuple[str, str, Relation, str, Optional[Tuple]]


class BeliefService:
    def __init__(
        self,
        statement_service: StatementServiceBase,
        notion_service: NotionService,
        categorical_deduction_service=None,
        conditional_deduction_service=None,
    ):
        if statement_service is None:
            raise ValueError("statement_service is required")
        self.statement_service = statement_service
        self.notion_service = notion_service
        self._categorical_deduction_service = categorical_deduction_service
        self._conditional_deduction_service = conditional_deduction_service
        self._matcher = ModifierMatcher()
        self._max_deduction_depth = 5

    def set_categorical_deduction_service(self, service) -> None:
        self._categorical_deduction_service = service

    def set_conditional_deduction_service(self, service) -> None:
        self._conditional_deduction_service = service

    def _time_key(self, time: Optional[Time]) -> Optional[Tuple[Optional[int], Optional[int], Optional[int], Optional[int], Optional[int]]]:
        if time is None:
            return None
        return (time.year, time.month, time.day, time.hour, time.minute)

    def _attributes_key(self, attributes: Optional[List[Attribute]]) -> Optional[Tuple[Tuple[Optional[str], Optional[float], Optional[str]], ...]]:
        if attributes is None:
            return None
        return tuple((attribute.tag, attribute.value, attribute.unit) for attribute in attributes)

    def _modifier_has_context(self, modifier: Optional[Modifier]) -> bool:
        if modifier is None:
            return False
        if modifier.location is not None:
            return True
        if modifier.from_time is not None:
            return True
        if modifier.to_time is not None:
            return True
        if modifier.deontic_modality is not None:
            return True
        if modifier.attributes:
            return True
        if modifier.quantity is not None:
            return True
        return False

    def _modifier_key(self, modifier: Optional[Modifier]) -> Optional[Tuple]:
        if not self._modifier_has_context(modifier):
            return None
        assert modifier is not None
        deontic = modifier.deontic_modality.value if modifier.deontic_modality is not None else None
        return (
            modifier.location,
            deontic,
            self._time_key(modifier.from_time),
            self._time_key(modifier.to_time),
            self._attributes_key(modifier.attributes),
            modifier.quantity,
        )

    def _apply_quantity_modifier(self, modifier: Optional[Modifier], quantity: Optional[float]) -> Optional[Modifier]:
        if quantity is None:
            return modifier
        if modifier is None:
            return Modifier(quantity=quantity)
        return Modifier(
            location=modifier.location,
            deontic_modality=modifier.deontic_modality,
            from_time=modifier.from_time,
            to_time=modifier.to_time,
            attributes=list(modifier.attributes) if modifier.attributes else None,
            quantity=quantity,
        )

    def _belief_key_from_statement(self, statement: StatementBase) -> BeliefKey:
        return self._belief_key_from_values(
            agent_id=statement.agent_id,
            subject_id=statement.subject_notion.id,
            object_id=statement.complement_notion.id,
            relation=statement.relation,
            modifier=statement.modifier,
        )

    def _belief_key_from_values(
        self,
        agent_id: str,
        subject_id: str,
        object_id: str,
        relation: Relation,
        modifier: Optional[Modifier],
    ) -> BeliefKey:
        if not isinstance(relation, Relation):
            raise TypeError("relation must be a Relation instance")
        return (
            agent_id,
            subject_id,
            relation,
            object_id,
            self._modifier_key(modifier),
        )

    def _modifier_matches(self, stored: Optional[Modifier], query: Optional[Modifier]) -> bool:
        stored_normalized = stored if self._modifier_has_context(stored) else None
        query_normalized = query if self._modifier_has_context(query) else None
        return self._matcher.context_matches(stored_normalized, query_normalized)

    def _modifier_exact_matches(self, stored: Optional[Modifier], query: Optional[Modifier]) -> bool:
        stored_normalized = stored if self._modifier_has_context(stored) else None
        query_normalized = query if self._modifier_has_context(query) else None
        return self._matcher.context_exact_matches(stored_normalized, query_normalized)

    def _resolve_notion_by_id(self, notion_id: str, agent_id: str) -> Notion:
        repository = getattr(self.notion_service, "notion_repository", None)
        if repository is not None:
            found = repository.find_by_id(notion_id)
            if found:
                return found
        return Notion(user_id="system", agent_id=agent_id, caption=notion_id, id=notion_id)

    def _build_belief(self, statements: Iterable[StatementBase], belief_id: Optional[str] = None) -> Belief:
        statements_list = list(statements)
        if not statements_list:
            raise ValueError("Cannot build belief without statements")
        ordered = sorted(
            statements_list,
            key=lambda stmt: getattr(stmt, "created_at", datetime.min),
        )
        first = ordered[0]
        belief = Belief(
            agent_id=first.agent_id,
            id=belief_id or first.id,
        )
        for statement in ordered:
            belief.add_statement(statement)
        created_candidates: List[datetime] = []
        for stmt in ordered:
            created_at = getattr(stmt, "created_at", None)
            if isinstance(created_at, datetime):
                created_candidates.append(created_at)
        created_at = min(created_candidates) if created_candidates else None
        updated_at = max(created_candidates) if created_candidates else None
        if created_at:
            belief.created_at = created_at
        if updated_at:
            belief.updated_at = updated_at
        return belief

    def _group_statements(self, statements: Iterable[StatementBase]) -> Dict[BeliefKey, List[StatementBase]]:
        groups: Dict[BeliefKey, List[StatementBase]] = {}
        for statement in statements:
            key = self._belief_key_from_statement(statement)
            groups.setdefault(key, []).append(statement)
        return groups

    def _base_beliefs_for_agent(
        self,
        agent_id: str,
        modifier: Optional[Modifier],
        quantity: Optional[float],
    ) -> Dict[BeliefKey, Belief]:
        modifier = self._apply_quantity_modifier(modifier, quantity)
        statements = self.statement_service.find_by_agent_id(agent_id)
        if modifier is None:
            filtered = list(statements)
        else:
            filtered = [
                stmt for stmt in statements
                if self._modifier_exact_matches(stmt.modifier, modifier)
            ]
        groups = self._group_statements(filtered)
        return {
            belief_key: self._build_belief(group)
            for belief_key, group in groups.items()
        }

    def _apply_deductions(
        self,
        beliefs_by_key: Dict[BeliefKey, Belief],
        agent_id: str,
        modifier: Optional[Modifier],
        quantity: Optional[float],
        user_id: Optional[str] = None,
    ) -> Dict[BeliefKey, Belief]:
        if self._categorical_deduction_service is not None:
            self._categorical_deduction_service.resolve_deductions(
                beliefs_by_key,
                modifier=modifier,
                max_depth=self._max_deduction_depth,
            )

        if self._conditional_deduction_service is not None:
            deductions = self._conditional_deduction_service.resolve_deductions(
                agent_id=agent_id,
                beliefs=list(beliefs_by_key.values()),
                modifier=modifier,
                user_id=user_id,
            )
            for deduction in deductions:
                belief_key = self._belief_key_from_statement(deduction)
                belief = beliefs_by_key.get(belief_key)
                if belief is None:
                    beliefs_by_key[belief_key] = self._build_belief([deduction])
                else:
                    belief.add_statement(deduction)

        return beliefs_by_key

    def _beliefs_for_agent(
        self,
        agent_id: str,
        modifier: Optional[Modifier],
        quantity: Optional[float],
        user_id: Optional[str] = None,
    ) -> Dict[BeliefKey, Belief]:
        beliefs_by_key = self._base_beliefs_for_agent(agent_id, modifier, quantity)
        self._apply_deductions(
            beliefs_by_key,
            agent_id=agent_id,
            modifier=modifier,
            quantity=quantity,
            user_id=user_id,
        )
        return beliefs_by_key

    def get_base_beliefs_for_agent(
        self,
        agent_id: str,
        modifier: Optional[Modifier] = None,
        quantity: Optional[float] = None,
    ) -> List[Belief]:
        beliefs_by_key = self._base_beliefs_for_agent(agent_id, modifier, quantity)
        return list(beliefs_by_key.values())


    @typechecked
    def add_statement_to_belief(self, statement: DirectStatement) -> Belief:
        statements = [statement]
        return self._build_belief(statements)

    @typechecked
    def get_recently_updated(self, agent_id: str, since: Optional[datetime] = None) -> List[Belief]:
        if since is None:
            since = datetime.now() - timedelta(seconds=20)
        beliefs = self.get_all_for_agent(agent_id)
        return [belief for belief in beliefs if belief.updated_at >= since or belief.created_at >= since]

    @typechecked
    def find(self, subject_notion: Notion, complement_notion: Notion,
             relation: Relation,
             modifier: Optional[Modifier] = None,
             quantity: Optional[float] = None,
             user_id: Optional[str] = None) -> Optional[Belief]:
        modifier = self._apply_quantity_modifier(modifier, quantity)
        belief_key = self._belief_key_from_values(
            agent_id=subject_notion.agent_id,
            subject_id=subject_notion.id,
            object_id=complement_notion.id,
            relation=relation,
            modifier=modifier,
        )
        beliefs_by_key = self._beliefs_for_agent(
            agent_id=subject_notion.agent_id,
            modifier=modifier,
            quantity=quantity,
            user_id=user_id,
        )
        belief = beliefs_by_key.get(belief_key)
        if belief is None:
            return None
        serialized = serialize_statement(
            subject_notion_caption=belief.subject_notion.caption,
            complement_notion_caption=belief.complement_notion.caption,
            relation=belief.relation,
            probability=float(belief.probability) if belief.probability is not None else 1.0,
            modifier=belief.modifier,
            subject_attributes=belief.subject_notion.attributes,
            complement_attributes=belief.complement_notion.attributes,
        )
        app_logger.debug(f"Found Belief: {serialized} (belief_id: {belief.id})")
        return belief

    @typechecked
    def find_by_expressions(self, agent_id: str, subject_notion_expression: NotionExpression,
                           complement_notion_expression: NotionExpression,
                           relation: Relation,
                           modifier: Optional[Modifier] = None,
                           user_id: Optional[str] = None) -> Optional[Belief]:
        subject_notion = self.notion_service.find(agent_id, subject_notion_expression)
        if not subject_notion:
            app_logger.warning(f"Source notion could not be found using expression '{subject_notion_expression.caption}'")
            return None
        complement_notion = self.notion_service.find(agent_id, complement_notion_expression)
        if not complement_notion:
            app_logger.warning(f"Target notion could not be found using expression '{complement_notion_expression.caption}'")
            return None
        return self.find(
            subject_notion=subject_notion,
            complement_notion=complement_notion,
            relation=relation,
            modifier=modifier,
            user_id=user_id,
        )

    @typechecked
    def search_by_notions(self, subject_notion: Notion, complement_notion: Notion,
                          modifier: Optional[Modifier] = None) -> List[Belief]:
        statements = self.statement_service.find_by_subject(subject_notion)
        filtered = [
            stmt for stmt in statements
            if stmt.complement_notion.id == complement_notion.id
            and self._modifier_matches(stmt.modifier, modifier)
        ]
        groups = self._group_statements(filtered)
        beliefs = [self._build_belief(group) for group in groups.values()]
        beliefs.sort(key=lambda belief: (belief.updated_at, belief.id), reverse=True)
        app_logger.debug(f"Found {len(beliefs)} beliefs - subject='{subject_notion.caption}', object='{complement_notion.caption}'")
        return beliefs

    @typechecked
    def search_by_subject_notion(self, subject_notion: Notion,
                                modifier: Optional[Modifier] = None) -> List[Belief]:
        statements = self.statement_service.find_by_subject(subject_notion)
        filtered = [stmt for stmt in statements if self._modifier_matches(stmt.modifier, modifier)]
        groups = self._group_statements(filtered)
        beliefs = [self._build_belief(group) for group in groups.values()]
        beliefs.sort(key=lambda belief: (belief.updated_at, belief.id), reverse=True)
        app_logger.debug(f"Found {len(beliefs)} beliefs - subject='{subject_notion.caption}'")
        return beliefs

    @typechecked
    def search_by_complement_notion(self, complement_notion: Notion,
                                modifier: Optional[Modifier] = None) -> List[Belief]:
        statements = self.statement_service.find_by_object(complement_notion)
        filtered = [stmt for stmt in statements if self._modifier_matches(stmt.modifier, modifier)]
        groups = self._group_statements(filtered)
        beliefs = [self._build_belief(group) for group in groups.values()]
        beliefs.sort(key=lambda belief: (belief.updated_at, belief.id), reverse=True)
        app_logger.debug(f"Found {len(beliefs)} beliefs - object='{complement_notion.caption}'")
        return beliefs

    @typechecked
    def search_by_subject_notion_and_relation(self, subject_notion: Notion,
                                                       relation: Relation,
                                                       modifier: Optional[Modifier] = None) -> List[Belief]:
        beliefs = self.search_by_subject_notion(subject_notion=subject_notion, modifier=modifier)
        return [b for b in beliefs if b.relation == relation]

    @typechecked
    def search_by_complement_notion_and_relation(self, complement_notion: Notion,
                                                      relation: Relation,
                                                      modifier: Optional[Modifier] = None) -> List[Belief]:
        beliefs = self.search_by_complement_notion(complement_notion=complement_notion, modifier=modifier)
        return [b for b in beliefs if b.relation == relation]

    @typechecked
    def search_by_subject_notion_expression(self, agent_id: str, subject_notion_expression: NotionExpression,
                                            modifier: Optional[Modifier] = None) -> Optional[List[Belief]]:
        subject_notion = self.notion_service.find(agent_id, subject_notion_expression)
        if not subject_notion:
            return None
        return self.search_by_subject_notion(subject_notion=subject_notion, modifier=modifier)

    @typechecked
    def search_by_complement_notion_expression(self, agent_id: str, complement_notion_expression: NotionExpression,
                                          modifier: Optional[Modifier] = None) -> Optional[List[Belief]]:
        complement_notion = self.notion_service.find(agent_id, complement_notion_expression)
        if not complement_notion:
            return None
        return self.search_by_complement_notion(complement_notion=complement_notion, modifier=modifier)

    @typechecked
    def forget(self, belief: Belief) -> None:
        serialized = serialize_statement(
            subject_notion_caption=belief.subject_notion.caption,
            complement_notion_caption=belief.complement_notion.caption,
            relation=belief.relation,
            probability=float(belief.probability) if belief.probability is not None else 1.0,
            modifier=belief.modifier,
            subject_attributes=belief.subject_notion.attributes,
            complement_attributes=belief.complement_notion.attributes,
        )
        app_logger.debug(f"Skipped persistence for Belief: {serialized} (belief_id: {belief.id})")

    @typechecked
    def get_all_for_agent(
        self,
        agent_id: str,
        user_id: Optional[str] = None,
    ) -> List[Belief]:
        beliefs_by_key = self._beliefs_for_agent(
            agent_id=agent_id,
            modifier=None,
            quantity=None,
            user_id=user_id,
        )
        beliefs = list(beliefs_by_key.values())
        app_logger.debug(f"Found {len(beliefs)} beliefs")
        return beliefs


    @typechecked
    def search_by_criteria(self, agent_id: str, clause: Dict["Criterion", Any]) -> bool:
        from analogai.foundation.extensions.belief_query.belief_criteria_matcher import BeliefCriteriaMatcher
        beliefs = self.get_all_for_agent(agent_id)
        return any(BeliefCriteriaMatcher.matches_clause(belief, clause) for belief in beliefs)

    @typechecked
    def search_by_modifier(self, modifier: Modifier) -> List[Belief]:
        app_logger.warning("search_by_modifier is not supported without agent scope in live mode")
        return []


from .statement_serializers import StatementSerializer, Verbalization, serialize_statement
from .expressions_serializer import (
    serialize_belief_expression,
    serialize_categorical_deduction_expression,
    serialize_hypothetical_statement_expression,
)
from .expression_factory import ExpressionFactory

__all__ = [
    'StatementSerializer',
    'Verbalization',
    'serialize_statement',
    'serialize_belief_expression',
    'serialize_categorical_deduction_expression',
    'serialize_hypothetical_statement_expression',
    'ExpressionFactory',
]



from analogai.foundation.extensions.conversation.services.conversation_service import ConversationService
from analogai.foundation.extensions.conversation.repositories import InMemoryConversationRepositoryImpl, MongoDBConversationRepositoryImpl
from analogai.foundation import config
from analogai.foundation.common.enums.environment import Environment

class ConversationServiceFactory:
    def create(self) -> ConversationService:
        env = Environment(config.ENVIRONMENT)
        if env == Environment.PRODUCTION:
            repository = MongoDBConversationRepositoryImpl()
        elif env == Environment.DEVELOPMENT:
            repository = InMemoryConversationRepositoryImpl()
        else:
            raise ValueError(f"No repository implementation for environment: {env}")
        return ConversationService(repository)



from __future__ import annotations

from typing import Optional
from fastapi import HTTPException, Request, status
from analogai.foundation.modules.user.user import User
from analogai.foundation.modules.user.user_service import UserService
from analogai.foundation.extensions.authentication.interfaces.authentication_factory_interface import (
    AuthenticationDependency,
    AuthenticationFactoryInterface,
    TokenExtractionDependency,
    ApiTokenAuthenticationDependency,
)
from analogai.foundation.extensions.authentication.core import IJWTService
from analogai.foundation.extensions.authentication.models import TokenPayload
from analogai.foundation.common.logging.app_logger import app_logger
from analogai.foundation.extensions.auth.authentication.api_token.api_token_authentication_service import (
    ApiTokenAuthenticationService,
)
from analogai.foundation.extensions.auth.authentication.api_token.auth_type_handlers import (
    ApiAuthTypeHandler,
    GuestApiAuthHandler,
    SystemApiAuthHandler,
    UserApiAuthHandler,
)

class AuthenticationFactory(AuthenticationFactoryInterface):
    def __init__(self):
        self._jwt_service: Optional[IJWTService] = None
        self._user_service: Optional[UserService] = None
        self._api_token_service: Optional[ApiTokenAuthenticationService] = None
        self._api_auth_handlers: Optional[dict[str, ApiAuthTypeHandler]] = None

    def create_authentication_dependency(self) -> AuthenticationDependency:
        async def dependency(request: Request) -> User:
            token = self._extract_token(request)
            try:
                payload = self._get_jwt_service().decode_payload(token)
            except ValueError as error:
                raise HTTPException(
                    status_code=status.HTTP_401_UNAUTHORIZED,
                    detail=str(error),
                ) from error

            if payload.get("email"):
                try:
                    user = self._get_jwt_service().authenticate_user(token)
                except ValueError as error:
                    detail = str(error)
                    if "user not found" in detail.lower():
                        detail = "User not found for token"
                    raise HTTPException(
                        status_code=status.HTTP_401_UNAUTHORIZED,
                        detail=detail,
                    ) from error
            else:
                user = await self._authenticate_api_token(request, token)

            if not user:
                raise HTTPException(
                    status_code=status.HTTP_401_UNAUTHORIZED,
                    detail="User not found for token",
                )

            return user

        return dependency

    def create_token_extraction_dependency(self) -> TokenExtractionDependency:
        async def dependency(request: Request) -> TokenPayload:
            token = self._extract_token(request)
            try:
                return self._get_jwt_service().extract_payload(token)
            except ValueError as error:
                raise HTTPException(
                    status_code=status.HTTP_401_UNAUTHORIZED,
                    detail=str(error),
                ) from error

        return dependency

    def create_api_token_authentication_dependency(self) -> ApiTokenAuthenticationDependency:
        async def dependency(request: Request) -> User:
            token = self._extract_token(request)
            return await self._authenticate_api_token(request, token)

        return dependency

    async def _authenticate_api_token(self, request: Request, token: str) -> User:
        try:
            api_payload = self._get_api_token_service().extract_payload(token)
        except ValueError as error:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail=str(error),
            ) from error

        auth_payload = await self._extract_api_auth_payload(request)
        auth_type = str(auth_payload.get("auth_type", "system")).lower()

        handler = self._get_api_auth_handlers().get(auth_type)
        if not handler:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Invalid auth_type; expected 'user', 'system', or 'guest'",
            )

        try:
            user = handler.authenticate(
                auth_payload=auth_payload,
                api_payload=api_payload,
                user_service=self._get_user_service(),
            )
        except ValueError as error:
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail=str(error),
            ) from error

        request.state.api_auth_type = auth_type
        request.state.api_agent_id = api_payload.agent_id
        request.state.api_user_email = user.email if user else None
        return user

    def _get_jwt_service(self) -> IJWTService:
        if not self._jwt_service:
            from analogai.foundation.extensions.authentication import authentication_factory as auth_factory

            self._jwt_service = auth_factory.Auth0JWTService(self._get_user_service())
        return self._jwt_service

    def _get_user_service(self) -> UserService:
        if not self._user_service:
            from analogai.foundation.extensions.authentication import authentication_factory as auth_factory

            repository = auth_factory.create_repository_strategy().get_user_repository()
            self._user_service = UserService(repository)
        return self._user_service

    def _get_api_token_service(self) -> ApiTokenAuthenticationService:
        if not self._api_token_service:
            self._api_token_service = ApiTokenAuthenticationService()
        return self._api_token_service

    def _get_api_auth_handlers(self) -> dict[str, ApiAuthTypeHandler]:
        if not self._api_auth_handlers:
            handlers = (
                UserApiAuthHandler(),
                SystemApiAuthHandler(),
                GuestApiAuthHandler(),
            )
            self._api_auth_handlers = {handler.auth_type: handler for handler in handlers}
        return self._api_auth_handlers

    async def _extract_api_auth_payload(self, request: Request) -> dict:
        try:
            payload = await request.json()
        except Exception:
            return {}
        if isinstance(payload, dict):
            return payload
        return {}


    def _extract_token(self, request: Request) -> str:
        authorization = request.headers.get("Authorization", "")
        if not authorization.startswith("Bearer "):
            app_logger.debug(
                "Invalid Authorization header format: '%s'",
                authorization,
            )
            raise HTTPException(
                status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Invalid authentication token format",
            )
        return authorization.replace("Bearer ", "", 1)



from typing import Optional

from analogai.foundation.extensions.belief_query.services.belief_query_service import BeliefQueryService
from analogai.foundation.modules.belief.belief_service import BeliefService
from analogai.foundation.common.utils.similarity.similarity_service import SimilarityService
from analogai.foundation import config


class BeliefQueryServiceFactory:
    def __init__(
        self,
        belief_service: BeliefService,
        similarity_service: Optional[SimilarityService] = None,
    ):
        self._belief_service = belief_service
        self._similarity_service = similarity_service

    def get_service(self) -> BeliefQueryService:
        similarity_service = self._similarity_service
        if similarity_service is None and config.USE_VECTOR_SEARCH:
            from analogai.foundation.infrastructure.services.embedding_similarity_service import (
                EmbeddingSimilarityService,
            )

            similarity_service = EmbeddingSimilarityService()
        return BeliefQueryService(
            belief_service=self._belief_service,
            similarity_service=similarity_service,
        )
